@@ -19,7 +19,7 @@ use BerkeleyDB qw(DB_CREATE DB_GET_BOTH DB_INIT_CDB DB_INIT_LOCK
                   DB_INIT_LOG DB_INIT_TXN DB_INIT_MPOOL DB_NEXT
                   DB_SET_RANGE DB_GET_BOTH DB_GET_BOTH_RANGE
                   DB_RECOVER DB_RECOVER_FATAL DB_DONOTINDEX
-                  DB_DUP DB_DUPSORT DB_DIRTY_READ DB_AUTO_COMMIT DB_HASH);
+                  DB_DUP DB_DUPSORT DB_DIRTY_READ DB_AUTO_COMMIT);
 
 use Path::Class  ();
 use File::Copy   ();
@@ -32,7 +32,6 @@ use List::Util   ();
 # directories
 use constant STORE   => 'store';
 use constant TEMP    => 'tmp';
-use constant CONTROL => 'control';
 use constant INDEX   => 'index';
 use constant BUFSIZE => 2**13;
 
@@ -114,11 +113,11 @@ Store::Digest::Driver::FileSystem - File system driver for Store::Digest
 
 =head1 VERSION
 
-Version 0.05
+Version 0.04
 
 =cut
 
-our $VERSION = '0.05';
+our $VERSION = '0.04';
 
 # target directory
 has dir => (
@@ -145,7 +144,7 @@ has _env => (
 # metadata for the actual database
 has _control => (
     is  => 'rw',
-    isa => 'BerkeleyDB::Btree',
+    isa => 'BerkeleyDB::Hash',
 );
 
 # the payload data itself
@@ -201,84 +200,6 @@ sub _create_dirs {
     }
 }
 
-# attempt to create a btree control file; rescue from the stupid
-# stupid stupid stupid idea that this should be a hash
-sub _init_control {
-    my $self = shift;
-
-    my $env  = $self->_env;
-    my $mode = 0666 & ~$self->umask;
-    my $file = $self->dir->absolute->file(CONTROL);
-
-    my $control;
-
-    if (-e $file) {
-        $control = BerkeleyDB::Unknown->new(
-            -Env      => $env,
-            -Flags    => DB_AUTO_COMMIT,
-            -Mode     => $mode,
-            -Filename => CONTROL,
-        ) or Carp::croak
-            ("Could not initialize control database: $BerkeleyDB::Error");
-
-        if ($control->type == DB_HASH) {
-            my $c2 = BerkeleyDB::Btree->new(
-                -Env      => $env,
-                -Flags    => DB_CREATE|DB_AUTO_COMMIT,
-                -Mode     => $mode,
-                -Filename => CONTROL . 2,
-            ) or Carp::croak
-                ("Could not create replacement control: $BerkeleyDB::Error");
-
-            my $txn = $env->txn_begin;
-            $txn->Txn($control, $c2);
-
-            my $cur = $control->db_cursor;
-            my ($k, $v, $ret) = ('', '', 0);
-            while ($cur->c_get($k, $v, DB_NEXT) == 0) {
-                #warn "wat $k $v";
-                $ret += $c2->db_put($k, $v);
-            }
-            Carp::croak("Could not copy database: $BerkeleyDB::Error") if $ret;
-            undef $cur;
-
-            unless ($txn->txn_commit == 0) {
-                die 'Could not copy contents from old control database to new!';
-            }
-
-            $env->txn_checkpoint(0, 0);
-
-            # require Data::Dumper;
-            # warn Data::Dumper::Dumper($env->txn_stat);
-
-            undef $txn;
-
-            $control->db_close;
-            $c2->db_close;
-            undef $control;
-            undef $c2;
-
-            # note OK is zero hence `and`
-            BerkeleyDB::db_remove(-Env => $env, -Filename => CONTROL)
-                  and die 'Could not delete old hash-based control database!: '
-                      . $BerkeleyDB::Error;
-            BerkeleyDB::db_rename
-                  (-Env => $env, -Filename => CONTROL . 2, -Newname => CONTROL)
-                      and die 'Could not rename new B+Tree control database!: '
-                          . $BerkeleyDB::Error;
-        }
-    }
-
-    $control ||= BerkeleyDB::Btree->new(
-        -Env      => $env,
-        -Flags    => DB_CREATE|DB_AUTO_COMMIT,
-        -Mode     => $mode,
-        -Filename => CONTROL,
-    ) or Carp::croak
-        ("Could not initialize (B+Tree) control database: $BerkeleyDB::Error");
-
-    $control;
-}
 
 sub BUILD {
     my $self = shift;
@@ -300,22 +221,23 @@ sub BUILD {
 
     $self->_env($env);
 
-    # create control
-    # my $control = BerkeleyDB::Hash->new(
-    #     -Env      => $env,
-    #     -Flags    => DB_CREATE|DB_AUTO_COMMIT,
-    #     -Mode     => $mode,
-    #     #-Txn      => $txn,
-    #     -Filename => 'control',
-    # ) or Carp::croak
-    #     ("Can't create/bind control database: $BerkeleyDB::Error");
-    my $control = $self->_init_control;
-    $self->_control($control);
-
     my $txn = $env->txn_begin;
+
+    # create control
+    my $control = BerkeleyDB::Hash->new(
+        -Env      => $env,
+        -Flags    => DB_CREATE|DB_AUTO_COMMIT,
+        -Mode     => $mode,
+        #-Txn      => $txn,
+        -Filename => 'control',
+    ) or Carp::croak
+        ("Can't create/bind control database: $BerkeleyDB::Error");
+
     # add to transaction
     $control->Txn($txn);
 
+    $self->_control($control);
+
     # control stores:
 
     # algorithms used
@@ -351,10 +273,7 @@ sub BUILD {
         my $v = join ',', @$w;
 
         $self->_algorithms($w);
-        unless ($control->db_put(algorithms => $v) == 0) {
-            die 'INTERNAL ERROR: could not store control data (algorithms): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(algorithms => $v);
     }
 
     # primary algorithm
@@ -378,61 +297,35 @@ sub BUILD {
                   unless grep { $_ eq $v } @{$self->_algorithms};
 
         $self->_primary($v);
-        unless ($control->db_put(primary => $v) == 0) {
-            die 'INTERNAL ERROR: could not store control data (primary): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(primary => $v);
     }
 
     # total number of known hash entries
     unless ($control->db_get(objects => my $v) == 0) {
-        unless ($control->db_put(objects => 0) == 0) {
-            die 'INTERNAL ERROR: could not store control data (objects): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(objects => 0);
     }
 
     # total number of hash entries where the payloads were removed
     unless ($control->db_get(deleted => my $v) == 0) {
-        unless ($control->db_put(deleted => 0) == 0) {
-            die 'INTERNAL ERROR: could not store control data (deleted): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(deleted => 0);
     }
 
     # total number of bytes stored
     unless ($control->db_get(bytes => my $v) == 0) {
-        unless ($control->db_put(bytes => 0) == 0) {
-            die 'INTERNAL ERROR: could not store control data (bytes): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(bytes => 0);
     }
 
 
     # store-wide creation and modification times
     my $now = time;
     unless ($control->db_get(ctime => my $v) == 0) {
-        unless ($control->db_put(ctime => $now) == 0) {
-            die 'INTERNAL ERROR: could not store control data (ctime): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(ctime => $now);
     }
 
     unless ($control->db_get(mtime => my $v) == 0) {
-        unless ($control->db_put(mtime => $now) == 0) {
-            die 'INTERNAL ERROR: could not store control data (mtime): ' .
-                $BerkeleyDB::Error;
-        }
+        $control->db_put(mtime => $now);
     }
 
-    unless ($txn->txn_commit == 0) {
-        die 'INTERNAL ERROR: Could not commit transaction on control data: ' .
-            $BerkeleyDB::Error;
-    }
-
-    # renew transaction
-    $txn = $env->txn_begin;
-
     # create algo btrees, these enable partial matches
     my $pri  = $self->_primary;
     my $ent  = $self->_entries;
@@ -550,8 +443,6 @@ sub DEMOLISH {
 
 =head2 add
 
-Add an object to the store.
-
 =cut
 
 # # File::MimeInfo::mimetype_isa only tells you if the child type is an
@@ -702,11 +593,11 @@ sub add {
     my ($self, %p) = @_;
     my $pri = $self->_primary;
     my $ctl = $self->_control;
-    my $env = $self->_env;
 
     # open transactions on all databases
-    my $txn = $env->txn_begin or Carp::croak
-        ("Could not open transaction: $BerkeleyDB::Error " . $env->status);
+    my $txn = $self->_env->txn_begin or Carp::croak
+        ("Could not open transaction: $BerkeleyDB::Error "
+             . $self->_env->status);
     $txn->Txn($ctl, values %{$self->_entries});
 
     # Step 1: record digests as we read the content handle into a
@@ -767,7 +658,7 @@ sub add {
         File::Copy::move($tempname, $target);
     }
 
-    # Step 3: alter database entries
+    # Step 2: alter database entries
 
     my $db  = $self->_entries->{$pri};
     my $now = time;
@@ -777,7 +668,6 @@ sub add {
     my $rec = '';
     utf8::downgrade($rec);
     $db->db_get($bin, $rec);
-
     if ($rec) {
         # create an object because it's easier to deal with
         $obj = $self->_inflate($bin, $rec);
@@ -788,36 +678,35 @@ sub add {
         if ($obj->dtime and $obj->dtime->epoch > 0) {
             $obj->dtime(undef);
 
-            my $nok = 0;
-
             # don't forget to decrement the deleted count by 1
             $ctl->db_get(deleted => my $deleted);
-            $nok += $ctl->db_put
-                (deleted => ($deleted && $deleted > 0 ? $deleted - 1 : 0));
+            $ctl->db_put(deleted =>  --$deleted) if $deleted > 0;
             # and no matter what, we don't want this number to be negative
 
             # set the modification time as well
-            $nok += $ctl->db_put(mtime => $now);
-
-            if ($nok) {
-                $txn->txn_abort;
-                die 'INTERNAL ERROR: Could not modify control database: ' .
-                    $BerkeleyDB::Error;
-            }
+            $ctl->db_put(mtime   => $now);
 
             # overwrite the record
             $rec = $self->_deflate($obj);
-
             # XXX ???? WTF how many times do i have to re-downgrade
             utf8::downgrade($rec);
+            #warn utf8::is_utf8($rec);
 
-            unless ($db->db_put($bin, $rec) == 0) {
-                die 'INTERNAL ERROR: Could not modify digest database: ' .
-                    $BerkeleyDB::Error;
-            }
+            $db->db_put($bin, $rec);
+            #$db->db_sync;
         }
     }
     else {
+        # set the mappings for the other digest algorithms
+        # my $e = $self->_entries;
+        # for my $algo (grep { $_ ne $pri } @{$self->_algorithms}) {
+        #     my $k = $p{digests}{$algo}->digest;
+        #     utf8::downgrade($k); # dat utf8
+
+        #     $e->{$algo}->db_put($k, $bin);
+        #     #$e->{$algo}->db_sync;
+        # }
+
         # tie up loose ends re the file
         my $stat    = $target->stat;
         $p{size}    = $stat->size;
@@ -832,43 +721,32 @@ sub add {
         $p{ctime} = $now;
 
         # now for the control db
-        my $nok = 0;
 
         # increment byte count by file size
         $ctl->db_get(bytes   => my $bytes);
-        $nok += $ctl->db_put(bytes   => ($bytes || 0) + $p{size});
+        $ctl->db_put(bytes   => $bytes += $p{size});
         # increment object count by 1
         $ctl->db_get(objects => my $objects);
-        $nok += $ctl->db_put(objects => ($objects || 0) + 1);
+        $ctl->db_put(objects =>  ++$objects);
         # set the modification time
-        $nok += $ctl->db_put(mtime   => $now);
-
-        if ($nok) {
-            $txn->txn_abort;
-            die 'INTERNAL ERROR: Could not modify control database: ' .
-                $BerkeleyDB::Error;
-        }
+        $ctl->db_put(mtime   => $now);
 
         # create the new object
         $obj = Store::Digest::Object->new(%p);
 
         # overwrite the record
         $rec = $self->_deflate($obj);
-
         # XXX ???? WTF
         utf8::downgrade($rec);
-
-        unless ($db->db_put($bin, $rec) == 0) {
-            die 'INTERNAL ERROR: Could not modify digest database: ' .
-                $BerkeleyDB::Error;
-        }
+        #warn utf8::is_utf8($rec);
+        $db->db_put($bin, $rec);
+        #$db->db_sync;
     }
 
     # Step 4: commit the changes
 
     unless ($txn->txn_commit == 0) {
-        die 'INTERNAL ERROR: Could not commit transaction: ' .
-            $BerkeleyDB::Error;
+        Carp::croak($BerkeleyDB::Error);
     }
 
     # Step 5: return the object
@@ -916,9 +794,9 @@ sub get {
     my $bin  = $digest->digest;
     utf8::downgrade($bin); # JIC
 
-    my @out = $self->_get($bin, $algo) or return;
+    my @out = $self->_get($bin, $algo);
 
-    wantarray ? @out : $out[0];
+    wantarray ? @out : \@out;
 }
 
 sub _get {
@@ -1192,25 +1070,26 @@ sub stats {
 
     my %x;
     for my $k (qw(objects deleted bytes ctime mtime)) {
-        die "INTERNAL ERROR: Control database has failed to store $k"
-            unless $ctl->db_get($k, my $val) == 0;
+        $ctl->db_get($k, my $val);
         $x{$k} = $val;
     }
 
-    # fix these into datetime objects
+    # fix these
     $x{created}  = DateTime->from_epoch(epoch => delete $x{ctime});
     $x{modified} = DateTime->from_epoch(epoch => delete $x{mtime});
 
     Store::Digest::Stats->new(%x);
 }
 
+# beginning to think i should index by ctime/mtime/dtime/type/encoding
+
 =head1 AUTHOR
 
-Dorian Taylor, C<E<lt>dorian at cpan.orgE<gt>>
+Dorian Taylor, C<< <dorian at cpan.org> >>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2012-2018 Dorian Taylor.
+Copyright 2012 Dorian Taylor.
 
 Licensed under the Apache License, Version 2.0 (the "License"); you
 may not use this file except in compliance with the License. You may
